package com.ddignani.network;


import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.StringTokenizer;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.Intent;
import android.database.Cursor;
import android.os.AsyncTask;
import android.widget.Toast;

import com.ddignani.KeepInQActivity;
import com.ddignani.R;
import com.ddignani.Reservation;
import com.ddignani.ReservationsDbAdapter;

public class Facade {	
	
	public static final int SOCKET_TIME_OUT = 10 * 1000; //tempo di attesa prima di lanciare eccezione (in ms)
	
	private KeepInQActivity mContext;
	private String phoneId;
	private ReservationsDbAdapter dbAdapter;

	public Facade(KeepInQActivity context, ReservationsDbAdapter dbAdapter, String phoneId){
		this.mContext = context;
		this.phoneId = phoneId;
		this.dbAdapter = dbAdapter;
	}

	/**
	 * Crea una nuova Reservation dalla stringa Decoded e avvia un AsyncTask per comunicare al server l'intenzione di accodarsi
	 * @param decoded Stringa ottenuta dalla lettura del QR-code
	 */
	public void newReservation(String decoded) {
		StringTokenizer tok = new StringTokenizer(decoded, "|");
		if(tok.countTokens() != 4){
			// Manda una reservation fasulla all'AsyncTask
			Reservation badRes = null;
			new NewReservationTask().execute(badRes);
			return;
		}
		Reservation res = new Reservation(tok.nextToken(), tok.nextToken(), Integer.parseInt(tok.nextToken()), tok.nextToken());
		new NewReservationTask().execute(res);
	}
	
	/**
	 * AsyncTask per una nuova prenotazione
	 */
	private class NewReservationTask extends AsyncTask <Reservation, Void, Reservation>{
		@Override
		protected void onPreExecute(){
			// Mostra la progress-bar nel titolo
	        mContext.setProgressBarIndeterminateVisibility(true);
		}
		
		@Override
		protected Reservation doInBackground(Reservation... params) {
			Reservation newRes = params[0];
			if(newRes == null){
				// E' la reservation fasulla inviata a seguito di un qr-code mal formattato
				newRes = new Reservation(null, null, 0, null);
				newRes.setStatus(Reservation.ERROR);
				newRes.setMsgFromServer("Bad QR-code format...");
				return newRes;
			}
			//Stabilisce la connessione
			openConn(newRes.getAddress(), newRes.getPort(), SOCKET_TIME_OUT);
			if(isItOpen()){
				ClientRequest request = new ClientRequest(phoneId, newRes.getPassword(), ClientRequest.NEW_RESERVATION);
				ServerResponse response = new ServerResponse();
				try{
					sendToServer(request);
					response = getFromServer();
				}catch (Exception e) {}
				
				switch(response.getStatus()){
					case ServerResponse.ERROR:
						newRes.setStatus(Reservation.ERROR);
						break;
						
					case ServerResponse.CONFIRMED:
						newRes.setStatus(Reservation.CONFIRMED);
						break;
						
					case ServerResponse.ENQUEUED:
						newRes.setWaiting(response.getTime());
						newRes.setStatus(Reservation.PENDING);
						break;
					
					case ServerResponse.READY:
						newRes.setWaiting(response.getTime());
						newRes.setStatus(Reservation.READY);
						break;
				}
				
				newRes.setMsgFromServer(response.getMoreInfo());
				newRes.setLastUpdate();
				closeAll();
			}else{
				newRes.setStatus(Reservation.ERROR);
				newRes.setMsgFromServer("No internet connection available");
			}
			return newRes;
		}
		
		@Override
		protected void onPostExecute(Reservation newRes){
			// Nasconde la progress-bar nel titolo
			mContext.setProgressBarIndeterminateVisibility(false);
			
			switch (newRes.getStatus()){
				case Reservation.CONFIRMED:
				case Reservation.READY:	// Caso possibile se al momento della registrazione mancano meno di 3 minuti al proprio turno...
					long id = dbAdapter.getIdFromResevation(newRes);
					Reservation res = dbAdapter.fetchReservation(id);
					res.setStatus(newRes.getStatus());
					res.setMsgFromServer(newRes.getMsgFromServer());
					dbAdapter.updateReservation(id, res);
					mContext.fillData();
					break;
			 
				case Reservation.ERROR:
					Toast.makeText(mContext, newRes.getMsgFromServer(), Toast.LENGTH_LONG).show();
					break;
					
				case Reservation.PENDING:
					dbAdapter.createReservation(newRes);
				 	mContext.fillData();
				 	break;
			}					
		}
	}
	
	
	/**
	 * Comunica al server che si sta rinunciando ad una reservation
	 * @param res Reservation a cui si sta rinunciando
	 */
	public void decline(long id) {
		new DeclineReservationTask().execute(id);
	}
	
	
	
	/**
	 * AsyncTastk per cancellare una prenotazione. Non importa se e come si ottiene risposta:  giusto una notifica al server
	 * e il client canceller comunque questa prenotazione...
	 */
	private class DeclineReservationTask extends AsyncTask <Long, Void, Long>{		
		@Override
		protected Long doInBackground(Long... params) {
			Reservation res = dbAdapter.fetchReservation(params[0]);
			if(res != null){
				//Stabilisce la connessione
				openConn(res.getAddress(), res.getPort(), SOCKET_TIME_OUT);
				if(isItOpen()){
					ClientRequest request = new ClientRequest(phoneId, res.getPassword(), ClientRequest.DECLINE_RESERVATION);
					try{
						sendToServer(request);
					}catch (Exception e) {}
				}
				closeAll();
			}
			return params[0];
		}
		
		protected void onPostExecute(Long  id){
			dbAdapter.removeReservation(id);
			mContext.fillData();	
		}
	}
	
	
	/**
	 * Comunica a tutti i server che si sta rinunciando alle reservation
	 */
	public void declineAll() {
		new DeclineAllReservationsTask().execute();
		
	}
	
	/**
	 * AsyncTastk per cancellare tutte le prenotazioni. Non importa se e come si ottiene risposta:  giusto una notifica al server
	 * e il client canceller comunque tutte le prenotazioni...
	 */
	private class DeclineAllReservationsTask extends AsyncTask <Void, Void, Void>{		
		@Override
		protected Void doInBackground(Void... params) {
			Cursor cursor = dbAdapter.fetchAllReservations();
			if(!cursor.moveToFirst()){
				//il database  vuoto
				cancel(true);
			}
			do{
				long id = cursor.getLong(cursor.getColumnIndex(ReservationsDbAdapter.KEY_ROWID));
				Reservation res = dbAdapter.fetchReservation(id);
				if(res != null){
					//Stabilisce la connessione
					openConn(res.getAddress(), res.getPort(), SOCKET_TIME_OUT);
					if(isItOpen()){
						ClientRequest request = new ClientRequest(phoneId, res.getPassword(), ClientRequest.DECLINE_RESERVATION);
						try{
							sendToServer(request);
						}catch (Exception e) {}
					}
					closeAll();
				}
			}while(cursor.moveToNext());
			cursor.close();
			return null;
		}
		
		protected void onPostExecute(Void v){
			dbAdapter.removeAllReservation();
			mContext.fillData();
		}
	}
	
	
	/**
	 * Chiede al server di aggiornare una Reservation
	 * @param id della Reservation da aggiornare
	 */
	public void updateReservation(long id) {
		new UpdateReservationTask().execute(id);
	}
	
	/**
	 * AsyncTask per aggiornare una Reservation
	 */
	private class UpdateReservationTask extends AsyncTask <Long, Void, Reservation>{

		protected void onPreExecute(){
			// Mostra la progress-bar nel titolo
	        mContext.setProgressBarIndeterminateVisibility(true);
		}
		
		@Override
		protected Reservation doInBackground(Long... params) {
			Reservation res = dbAdapter.fetchReservation(params[0]);
			if(res != null){
				//Stabilisce la connessione
				openConn(res.getAddress(), res.getPort(), SOCKET_TIME_OUT);
				if(isItOpen()){
					ClientRequest request = new ClientRequest(phoneId, res.getPassword(), ClientRequest.REFRESH_RESERVATION);
					ServerResponse response = new ServerResponse();
					try{
						sendToServer(request);
						response = getFromServer();
					}catch (Exception e) {
						response.setStatus(ServerResponse.ERROR);
					}

					if(response.getStatus() != ServerResponse.ERROR){
						res.setWaiting(response.getTime() - (System.currentTimeMillis() - res.getLastUpdate()));
						res.setStatus(Reservation.PENDING);
						res.setMsgFromServer(response.getMoreInfo());
						// Il server dice che bisogna prepararsi 
						if(response.getStatus() == ServerResponse.READY){
							res.setStatus(Reservation.READY);
							res.setMsgFromServer(response.getMoreInfo());
						}
					}else{
						res.setStatus(Reservation.ERROR);
						res.setMsgFromServer(response.getMoreInfo());
					}
					res.setLastUpdate();
					closeAll();
				}else{
					res.setStatus(Reservation.ERROR);
					res.setMsgFromServer("No internet connection available");
				}
			}
			return res;
		}
		
		protected void onPostExecute(Reservation res){
			// Nasconde la progress-bar nel titolo
			mContext.setProgressBarIndeterminateVisibility(false);
			
			if(res.getStatus() != Reservation.ERROR){
				long id = dbAdapter.getIdFromResevation(res);
				dbAdapter.updateReservation(id, res);
				mContext.fillData();				
			}
			else
				//Toast per dire che c' stato qualche errore a contattare il server
				Toast.makeText(mContext, res.getMsgFromServer(), Toast.LENGTH_LONG).show();
		}
	}
	
	
	/**
	 * Aggiorna tutte le Reservation (utilizza lo stesso AsyncTask dell'autoupdater)
	 */
	public void updateAllReservations() {
		new AutoUpdateReservationTask().execute();
		
	}
	
	/**
	 * Invocata dall'autoUpdater per aggiornare tutte le reservations (una per volta)
	 */
	public void autoUpdateReservations() {
		new AutoUpdateReservationTask().execute();
	}
	
	/**
	 * AsyncTask per aggiornare una Reservation tramite AutoUpdater
	 */
	private class AutoUpdateReservationTask extends AsyncTask <Void, Void, Integer>{
		
		@Override
		protected Integer doInBackground(Void... params) {
			final int NOTHING = 0;
			final int ALMOST_READY = 1;
			final int ERRORS = -1;
			
			int somethingImportant = NOTHING;
			
			Cursor cursor = dbAdapter.fetchAllReservations();
			if(!cursor.moveToFirst()){
				//il database  vuoto
				cancel(true);
			}
			do{
				long id = cursor.getLong(cursor.getColumnIndex(ReservationsDbAdapter.KEY_ROWID));
				Reservation res = dbAdapter.fetchReservation(id);
				if(res.getStatus() != Reservation.READY && res.getStatus() != Reservation.CONFIRMED){ // Se  gi Ready o Confirmed non la controllo ancora
					//Stabilisce la connessione
					openConn(res.getAddress(), res.getPort(), SOCKET_TIME_OUT);
					if(isItOpen()){
						ClientRequest request = new ClientRequest(phoneId, res.getPassword(), ClientRequest.REFRESH_RESERVATION);
						ServerResponse response = new ServerResponse();
						try{
							sendToServer(request);
							response = getFromServer();
						}catch (Exception e) {
							response.setStatus(ServerResponse.ERROR);
						}

						if(response.getStatus() != ServerResponse.ERROR){
							res.setWaiting(response.getTime() - (System.currentTimeMillis() - res.getLastUpdate()));
							res.setStatus(Reservation.PENDING);
							res.setMsgFromServer(response.getMoreInfo());
							// Il server dice che bisogna prepararsi o manca meno del prossimo autoUpdate
							if(response.getStatus() == ServerResponse.READY || response.getTime() <= mContext.getNextUpdateInterval()){
								res.setStatus(Reservation.READY);
								somethingImportant = ALMOST_READY;
							}
						}else{
							res.setStatus(Reservation.ERROR);
							res.setMsgFromServer(response.getMoreInfo());
							somethingImportant = ERRORS;
						}
						res.setLastUpdate();
						closeAll();
					}

					//Aggiorno il db
					dbAdapter.updateReservation(id, res);
				}
			}while(cursor.moveToNext());
			
			return somethingImportant;
		}
		
		protected void onPostExecute(Integer somethingImportant){
			//aggiorna la vista
			mContext.fillData();
			
			final int NOTHING = 0;
			final int ALMOST_READY = 1;
			final int ERRORS = -1;
			
			CharSequence ticker ="";
			CharSequence title ="";
			CharSequence message ="";
			
			switch(somethingImportant){
			case NOTHING:
				return;
			case ALMOST_READY:
				ticker = "KeepInQ";
				title = "Update";
				message = "It's almost your turn!";
				break;
			case ERRORS:
				ticker = "KeepInQ";
				title = "Warning";
				message = "Something went wrong...";
				break;
			}

			sendNotification(ticker, title, message);

		}
	}
	
	/**
	 * Crea e mostra una notifica che selezionata rimanda alla schermata principale di KeepInQ
	 * @param ticker Ticker della notifica
	 * @param title Titolo della notifica
	 * @param message Testo della notifica
	 */
	private void sendNotification(CharSequence ticker, CharSequence title,	CharSequence message) {
		// notifica
		int icon = R.drawable.ic_menu_refresh;

		Intent notificationIntent = new Intent(mContext, KeepInQActivity.class);
		notificationIntent.setAction(Intent.ACTION_MAIN);			//se dichiaro un activity come quella lanciata all'inizio, non ne crea un'altra
		notificationIntent.addCategory(Intent.CATEGORY_LAUNCHER);
		PendingIntent pendingIntent = PendingIntent.getActivity(mContext, 0, notificationIntent, 0);

		Notification notification = new Notification(icon, ticker, System.currentTimeMillis());
		notification.setLatestEventInfo(mContext, title, message, pendingIntent);
		notification.flags |= Notification.FLAG_AUTO_CANCEL;
		notification.defaults |= Notification.DEFAULT_SOUND;         
		// ;-)
		long DOT = 100, DASH = DOT * 3, GAP = DOT, LETTERGAP= DOT* 3;
		long vibe [] = new long [] {0, DASH, GAP, DOT, GAP, DASH, LETTERGAP, DOT, GAP, DOT, LETTERGAP, DASH, GAP, DASH, GAP, DOT, GAP, DASH}; 
		notification.vibrate = vibe;
		NotificationManager nm = (NotificationManager)mContext.getSystemService("notification");
		nm.notify(0, notification);

	}
	
	
	
/*///////////////////////////////////////////////////////
 * Funzioni di comunicazione col server
 *///////////////////////////////////////////////////////
	
	private Socket socket;
	private ObjectInputStream oIn;
	private ObjectOutputStream oOut;
	private boolean isItOpen = false;
	
	/**
	 * Stabilisce una connessione col server
	 * @param serverName Indirizzo o nome del server
	 * @param portNumber Porta di ascolto del server
	 * @param timeout Timeout della socket
	 */
	public void openConn(String serverName, int portNumber, int timeout){
		try{
			socket = new Socket(serverName, portNumber);
			socket.setSoTimeout(timeout);
			oOut = new ObjectOutputStream(socket.getOutputStream());
			oIn = new ObjectInputStream(socket.getInputStream());
			isItOpen = true;
		}catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Stabilisce se la connessione  stata stabilita o meno
	 * @return True se stabilita, False altrimenti
	 */
	public boolean isItOpen(){
		return isItOpen;
	}
	
	/**
	 * Legge una ServerResponse dal server
	 * @return La ServerResponse ricevuta dal server
	 * @throws IOException
	 */
	public ServerResponse getFromServer() throws IOException{
		ServerResponse response = null;

		try {
			response = (ServerResponse) oIn.readObject();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}

		if(response == null){
			response = new ServerResponse();
			response.setStatus(ServerResponse.ERROR);
		}
		return response;
	}
	
	
	/**
	 * Invia una ClientRequest al server
	 * @param request ClientRequest da inviare al server
	 * @throws IOException
	 */
	public void sendToServer(ClientRequest request) throws IOException{
		oOut.writeObject(request);
	}
	
	/**
	 * Chiude tutto. connessioni e stream
	 */
	public void closeAll(){
		try {
			socket.close();
			oIn.close();
			oOut.close();
			isItOpen = false;
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}





}

